/*****************************************************************************
 * Lemonade JS
 * https://lemonadejs.com
 *
 * ----
 *
 * Mysql
 *
 *  Adaptor to connect and execute queries on a Mysql replication farm. Will
 *  create different connections depending on the first word in the query,
 *  directing modifying requests to the master and selects to the slave
 *
 *****************************************************************************/

/**
 * Module dependencies
 */
var adaptor = require('mysql');

/**
 * Mysql
 *
 * @param {string} db
 * @param {object} master
 * @param {object} slave
 * @constructor
 */
function Mysql(db, master, slave) {

    /**
     * Status object
     */
    this.status = {
        querycnt : {
             master : 0
            ,slave  : 0
        }
        ,running : {}
    };

    /**
     * Master pool
     * @type {*}
     * @private
     */
    this._masterpool = adaptor.createPool({
         host     : master.host
        ,port     : master.port
        ,user     : master.user
        ,password : master.pass
        ,database : db

        ,connectionLimit : 10
    });

    if (typeof slave !== 'undefined') {
        /**
         * Slave pool
         * @type {*}
         * @private
         */
        this._slavepool = adaptor.createPool({
             host     : slave.host
            ,port     : slave.port
            ,user     : slave.user
            ,password : slave.pass
            ,database : db
    
            ,connectionLimit : 10
        });
    } else {
        this._slavepool = this._masterpool;
        this.status.singlepool = true;
    }

    /**
     * Wait until both pools are connected, then emit
     * a general connect event
     */
    this._poolsnc = 2;
    this.kernel.on('_Mysql-pool-connect', function() {
        if (!--this._poolsnc) {
            this.kernel.emit('Mysql-connect');
        }
    }.bind(this));

    /**
     * Try to connect the pools
     */
    this.connect();
}

/**
 * Do a query on each pool (master and slave), on both
 * queries success' we have a "connected" adaptor
 *
 * @return
 */
Mysql.prototype.connect = function() {

    this._poolsnc = 2;
    var cb = function(error) {
        if (error) {
            this.kernel.emit('Mysql-connect-error', error);
        } else {
            this.kernel.emit('_Mysql-pool-connect');
        }

    }.bind(this);

    /**
     * Dummy select on the slave
     */
    this.query('SELECT 1', [], cb);

    /**
     * Dummy select on the master
     */
    this.query('/* Force master */ SELECT 1', [], cb);

};

/**
 * Fetches a connection from the pool and runs a query
 *
 * @param {string} query
 * @param {array} params
 * @param {function} callback
 */
Mysql.prototype.query = function(query, params, callback) {
    var
    pool = choosePool.call(this, query);
    pool.getConnection(function(err, conn) {
        if (!err) {

            /**
             * Add the query to the status object
             * @type {*}
             */
            var qid = Math.floor(Math.random() * 1000);
            this.status.running[qid] = {
                 query  : query
                ,params : params
            };

            var execute = function(query, params, callback, qid) {

                /**
                 * Execute the query
                 */
                conn.query(query, params, function(err, results) {

                    if (err) {
                        callback(
                            new Error('Mysql query error : ' + err.message)
                        );
                    } else {
                        callback(null, results);
                    }

                    /**
                     * Release the connection
                     */
                    conn.end();

                    /**
                     * Remove the query from the status object
                     */
                    delete this.status.running[qid];

                }.bind(this));

            }.bind(this); execute(query, params, callback, qid);

        } else {
            callback(
                new Error('Mysql connection error : ' + err.message)
            );
        }
    }.bind(this));
};

/**
 * Chooses connection pool based on first word inside the query
 *
 * @param {string} query
 * @return {object}
 */
var choosePool = function(query) {
    var firstWord = query.substring(0, query.indexOf(' ')).toLowerCase();
    if (firstWord === 'select') {
        this.status.querycnt.slave++;
        return this._slavepool;
    }
    this.status.querycnt.master++;
    return this._masterpool;
};

/**
 * Exports
 */
exports = module.exports = Mysql;