/*****************************************************************************
 * Lemonade JS
 * https://lemonadejs.com
 *
 * ----
 *
 * Rabbit
 *
 *  Adaptor to connect to Rabbitmq. Subscribe to a queue and have messages
 *  pushed. Publish messages.
 *
 *****************************************************************************/

/**
 * Module dependencies
 */
var amqp = require('amqp');

/**
 * Rabbit
 * @constructor
 */
function Rabbit() {

    /**
     * Hosts to connect to
     * @type {Array}
     */
    this.host = [];

    /**
     * Internal flag for established connection
     * @type {boolean}
     * @private
     */
    this._connected  = false;

    /**
     * Internal flag for connecting in progress
     * @type {boolean}
     * @private
     */
    this._connecting = false;

    /**
     * Amqp connection object
     * @type {null}
     * @private
     */
    this._connection = null;

    /**
     * Prepared queues
     * @type {{}}
     * @private
     */
    this._prepared = {};

    /**
     * Number of reconnects
     * @type {number}
     * @private
     */
    this._reconnects = 0;

    /**
     * Max number of attempts to reconnect
     * @type {{}}
     * @private
     */
    this._maxreconnects = 3;

    /**
     * Subscribed queues
     * @type {{}}
     * @private
     */
    this._subscribed = {};

    /**
     * Status object
     * @returns {{}}
     */
    this.status = {
         subscribed  : this._subscribed
        ,published   : this._prepared
    };

    /**
     * Prefetch count
     * @type {number}
     * @private
     */
    this._prefetch = 10;
}

/**
 * Connect. Handles connection fails.
 * @return
 */
Rabbit.prototype.connect = function() {
    if (!this._connected && !this._connecting) {

        /**
         * Initiate the connection
         */
        connect.call(this);

        /**
         * Wait for ready signal. Emit custom event
         * when ready.
         */
        this._connection.on('ready', function() {

            this._reconnects = 0;
            this._connecting = false;
            this._connected  = true;

            this.kernel.emit('Rabbit-connected');

        }.bind(this));

        /**
         * Wait for connection closed. On this._maxreconnects
         * attempts throw up
         */
        this._connection.on('closed', function() {

            if (this._reconnects++ >= this._maxreconnects) {
                throw new Error('Rabbit connection lost and failed reconnecting');
            }

            this._connecting = false;
            this._connected  = false;
            this.connect();

        }.bind(this));

    }
};

/**
 * Initiate connection. An event will be emitted when ready.
 * @return
 */
function connect() {
    this._connecting = true;
    this._connection = amqp.createConnection({
         host : this.host
        ,defaultExchangeName : "" //use no exchange, bind directly to queues
    });
}

/**
 * Subscribe to a queue. Will receive messages as they are published. Instant ack.
 *
 * @param {string} queue
 * @param {function} callback
 * @return
 */
Rabbit.prototype.subscribe = function(queue, callback) {
    this._connection.queue(queue, {autoDelete : false, durable : true}, function(q) {

        /**
         * Mark queue as prepared (needed for publishing)
         */
        this._prepared[queue] = {
             queue     : queue
            ,published : 0
        };

        q.subscribe({ack : true, prefetchCount : this._prefetch}, function (message, headers, deliveryInfo) {

            /**
             * Increment no. of fetched messages for this queue
             */
            if (typeof this._subscribed[queue] !== 'undefined') {
                this._subscribed[queue].fetched++;
            }

            /**
             * If the message retrieved is and object, transform it to string
             */
            callback(null, (typeof message.data !== 'undefined') ? message.data.toString() : message);

            /**
             * Move to the next message
             */
            q.shift();

        }.bind(this)).addCallback(function(ok) {

            /**
             * Keep the consumer tag for debugging
             */
            if (typeof this._subscribed[queue] === 'undefined') {
                this._subscribed[queue] = {
                     queue   : queue
                    ,tag     : ok.consumerTag
                    ,active  : true
                    ,fetched : 1
                };
            } else {
                this._subscribed[queue].active = true;
                this._subscribed[queue].tag    = ok.consumerTag;
            }

        }.bind(this));
    }.bind(this));
};

/**
 * Unsubscribe from a queue
 *
 * @param {string} queue
 * @param {function} callback
 * @return
 */
Rabbit.prototype.unsubscribe = function(queue, callback) {
    if (typeof this._subscribed[queue] !== 'undefined') {
        this._connection.queue(queue, {autoDelete : false, durable : true}, function(q) {

            q.unsubscribe(this._subscribed[queue]);
            this._subscribed[queue].active = false;
            callback();

        }.bind(this));
    }
};

/**
 * Publish message to queue
 *
 * @param {string} queue
 * @param {string} message
 * @param {function} callback
 * @return
 */
Rabbit.prototype.publish = function(queue, message, callback) {
    if (typeof this._prepared[queue] === 'undefined') {
        /**
         * Initiate a connection to the queue first to create it if it's not
         * already there
         */
        this._connection.queue(queue, {autoDelete : false, durable : true}, function() {

            this._connection.publish(queue, message);
            this._prepared[queue] = {
                 queue     : queue
                ,published : 1
            };
            callback();

        }.bind(this));
    } else {

        this._connection.publish(queue, message);
        this._prepared[queue].published++;
        callback();

    }
};

/**
 * Export
 */
exports = module.exports = Rabbit;