/*****************************************************************************
 * Lemonade JS
 * https://lemonadejs.com
 *
 * ----
 *
 * Solr
 *
 *  Adaptor to connect and execute queries on a Apache Solr farm. Will direct
 *  add / delete doc requests to the master and searches to the slave.
 *
 *****************************************************************************/

/**
 * Module dependencies
 */
var solr = require('solr-client');

/**
 * Solr
 *
 * @constructor
 * @param {string} core
 * @param {object} master
 * @param {object} slave
 */
function Solr(core, master, slave) {

    /**
     * Status object
     */
    this.status = {
         ops : {
             add    : 0
            ,del    : 0
            ,search : 0
         }
        ,running : {}
    };

    /**
     * Master client
     * @type {*}
     * @private
     */
    this._master = solr.createClient(
          master.host
        , master.port
        , master.core || core
        , '/solr'
    );

    if (master !== slave) {
        /**
         * Slave client
         * @type {*}
         * @private
         */
        this._slave = solr.createClient(
              slave.host
            , slave.port
            , slave.core || core
            , '/solr'
        );
    } else {
        this._slave = this._master;
    }

    /**
     * Listen for con-ready events and emit
     * when both pools are connected
     */
    this._nc = 2;
    this.kernel.on('_Solr-con-ready', function() {
        if (!--this._nc) {
            this.kernel.emit('Solr-connect');
        }
    }.bind(this));

    /**
     * Try the connections
     */
    this.connect();
}

/**
 * Connect
 * @return
 */
Solr.prototype.connect = function() {

    /**
     * Reset unconnected count
     */
    this._nc = 2;

    /**
     * Build a test query
     */
    var q = this.query();
        q . q('test');
        q . rows(1);

    var cb = function(err, result) {
        if (err) {
            this.kernel.emit(
                 'Solr-connect-error'
                ,new Error('Solr connection error : ' + err)
            );
        } else {
            this.kernel.emit('_Solr-con-ready');
        }
    }.bind(this);

    /**
     * Test master client
     */
    this._master.search(q, cb);

    if (this._master !== this._slave) {
        /**
         * Test slave client
         */
        this._slave.search(q, cb);
    } else {
        this.kernel.emit('_Solr-con-ready');
    }
};

/**
 * Add document[s]
 *
 * @param {Array|Object} doc
 * @param {Function} callback
 * @return
 */
Solr.prototype.add = function(doc, callback) {
    this.status.ops.add++;
    this._master.add(prepare(doc), callback);
};

/**
 * Transform object to solr document
 *
 * @param {object} doc
 * @return {array}
 */
var prepare = function(doc) {
    var prepared = {}, key, ikey, iikey, date;
    for (key in doc) {

        /**
         * Transform date into ISO standard
         */
        if (key.indexOf('date') === 0) {
            if (typeof doc[key] === 'object') {
                for (ikey in doc[key]) {
                    date = new Date(doc[key][ikey] * 1000);
                    prepared[key + '_' + ikey] = date.toISOString();
                }
            } else {
                date = new Date(doc[key] * 1000);
                prepared[key] = date.toISOString();
            }
        }

        /**
         * Transform prices into solr currency type
         */
        else if (key.indexOf('price') === 0) {
            if (typeof doc[key] === 'object') {
                for (ikey in doc[key]) {
                    prepared[key + '_' + ikey] = doc[key][ikey].amount + ',' +
                        doc[key][ikey].currency.toUpperCase();
                }
            } else {
                prepared[key] = doc[key].amount + ',' +
                    doc[key].currency.toUpperCase();
            }
        }

        /**
         * Normal objects + multivalue
         */
        else if (typeof doc[key] === 'object') {
            for (ikey in doc[key]) {
                if (typeof doc[key][ikey] === 'object') {
                    var foo = [];
                    for (iikey in doc[key][ikey]) {
                        foo.push(doc[key][ikey][iikey]);
                    }
                    prepared[key + '_' + ikey] = foo;
                } else {
                    prepared[key + '_' + ikey] = doc[key][ikey];
                }
            }

        /**
         * Basic column => value
         */
        } else {
            prepared[key] = doc[key];
        }
    }

    return prepared;
};

/**
 * Delete by id
 *
 * @param {Number} id
 * @param {function} callback
 * @return
 */
Solr.prototype.delete = function(id, callback) {
    this.status.ops.del++;
    this._master.deleteByID(id, callback);
};

/**
 * Return a query object. Configure your query :
 *
 *     http://lbdremy.github.com/solr-node-client/code/query.js.html
 *
 * When ready, use search to execute the query
 *
 * @return {Query}
 */
Solr.prototype.query = function() {
    return this._slave.createQuery();
};

/**
 * Execute the query (Must have a query constructed using Solr.query())
 *
 * @param {Query} query
 * @param {Function} callback
 * @return
 */
Solr.prototype.search = function(query, callback) {
    this.status.ops.search++;
    this._slave.search(query, function(error, object) {
        callback(error, object);
    });
};

/**
 * Export
 */
exports = module.exports = Solr;